__author__ = 'Calvin Lobo'
__version__ = '1.81'

import contextlib
import json

from .property import Property
from .dictproperty import DictProperty, ObservableDict
from .limitproperty import LimitProperty
from .listproperty import ListProperty, ObservableList
from .optionproperty import OptionProperty
from .scheduledevent import ScheduledEvent
from .setproperty import SetProperty
from .stringproperty import StringProperty, _
from .unitproperty import UnitProperty
from .weakrefproperty import WeakRefProperty
from .exceptions import *


class EventDispatcher(object):
    def __init__(self, *args, **kwargs):
        self.event_dispatcher_event_callbacks = {}
        self.event_dispatcher_properties = {}
        bindings = {}
        # Walk backwards through the MRO looking for Property attributes in the classes. Then register and bind them to
        # 'on_<prop_name>' if it exists. Walking backwards allows you to override the default value for a superclass.
        for cls in reversed(self.__class__.__mro__):
            for prop_name, prop in cls.__dict__.iteritems():
                if isinstance(prop, Property):
                    prop.name = prop_name
                    prop.register(self, prop_name, prop.default_value)
                    if hasattr(self, 'on_{}'.format(prop_name)):
                        func = getattr(self, 'on_{}'.format(prop_name))
                        bindings.update({prop_name: func})

        self.bind(**bindings)

    def force_dispatch(self, prop_name, value):
        """
        Assigns the value to the property and then dispatches the event, regardless of whether that value is the same
        as the previous value.
        :param prop_name: property name
        :param value: value to assign to the property
        """
        previous_value = getattr(self, prop_name)
        if previous_value == value:
            self.dispatch(prop_name, self, previous_value)
        else:
            setattr(self, prop_name, value)

    def dispatch(self, key, *args, **kwargs):
        """
        Dispatch a property. This calls all functions bound to the property.
        :param event: property name
        :param args: arguments to provide to the bindings
        :param kwargs: keyword arguments to provide to the bindings
        """
        for callback in self.event_dispatcher_properties[key]['callbacks']:
            if callback(*args, **kwargs):
                break

    def dispatch_event(self, event, *args, **kwargs):
        """
        Dispatch an event. This calls all functions bound to the event.
        :param event: event name
        :param args: arguments to provide to the bindings
        :param kwargs: keyword arguments to provide to the bindings
        """
        for callback in self.event_dispatcher_event_callbacks[event]:
            if callback(*args, **kwargs):
                break

    def register_event(self, name):
        """
        Create an event that can be bound to and dispatched.
        :param name: Name of the event
        """
        if hasattr(self, 'on_{}'.format(name)):
            self.event_dispatcher_event_callbacks[name] = [getattr(self, 'on_{}'.format(name))]
        else:
            self.event_dispatcher_event_callbacks[name] = []

    def unbind(self, **kwargs):
        """
        Unbind the specified callbacks associated with the property / event names
        :param kwargs: {property name: callback} bindings
        """
        all_properties = self.event_dispatcher_properties
        for prop_name, callback in kwargs.iteritems():
            if prop_name in all_properties:
                try:
                    all_properties[prop_name]['callbacks'].remove(callback)
                except ValueError:
                    raise BindError("No binding for {} in property '{}'".format(callback.__name__, prop_name))
            elif prop_name in self.event_dispatcher_event_callbacks:
                try:
                    self.event_dispatcher_event_callbacks[prop_name].remove(callback)
                except ValueError:
                    raise BindError("No binding for {} in event '{}'".format(callback.__name__, prop_name))
            else:
                raise BindError('No property or event by the name of %s' % prop_name)

    def unbind_all(self, *args):
        """
        Unbind all callbacks associated with the specified property / event names
        :param args: property / event names
        """
        all_properties = self.event_dispatcher_properties
        for prop_name in args:
            if prop_name in all_properties:
                del all_properties[prop_name]['callbacks'][:]
            elif prop_name in self.event_dispatcher_event_callbacks:
                del self.event_dispatcher_event_callbacks[prop_name][:]
            else:
                raise BindError("No such property or event '%s'" % prop_name)

    def bind(self, **kwargs):
        """
        Bind a function to a property or event.
        :param kwargs: {property name: callback} bindings
        """
        for prop_name, callback in kwargs.iteritems():
            if prop_name in self.event_dispatcher_properties:
                # Queue the callback into the property
                self.event_dispatcher_properties[prop_name]['callbacks'].append(callback)
            elif prop_name in self.event_dispatcher_event_callbacks:
                # If a property was not found, search in events
                self.event_dispatcher_event_callbacks[prop_name].append(callback)
            else:
                raise BindError("No property or event by the name of '%s'" % prop_name)

    def bind_once(self, **kwargs):
        """
        Bind a function to a property or event and unbind it after the first time the function has been called
        :param kwargs: {property name: callback} bindings
        """
        for prop_name, callback in kwargs.items():
            def _wrapped_binding(*args):
                callback()
                self.unbind(**{prop_name: _wrapped_binding})

            self.bind(**{prop_name: _wrapped_binding})
            kwargs.pop(prop_name)
            if kwargs:
                self.bind_once(**kwargs)
                return

    def setter(self, prop_name):
        return lambda inst, value: setattr(self, prop_name, value)

    def get_dispatcher_property(self, prop_name):
        return self.event_dispatcher_properties[prop_name]['property']

    @contextlib.contextmanager
    def temp_unbind(self, **bindings):
        """
        Context manager to temporarily suspend dispatching of a specified callback.
        :param bindings: keyword argument of property_name=callback_func
        """
        # Enter / With
        all_properties = self.event_dispatcher_properties
        callbacks = {}
        for prop_name, binding in bindings.iteritems():
            if prop_name in all_properties:
                # Make a copy of the callback sequence so we can revert back
                callbacks[prop_name] = all_properties[prop_name]['callbacks'][:]
                # Remove the specified bindings
                if binding in all_properties[prop_name]['callbacks']:
                    all_properties[prop_name]['callbacks'].remove(binding)
            elif prop_name in self.event_dispatcher_event_callbacks:
                callbacks[prop_name] = self.event_dispatcher_event_callbacks[prop_name][:]
                self.event_dispatcher_event_callbacks[prop_name].remove(binding)
        # Inside of with statement
        yield None
        # Finally / Exit
        for prop_name, cb in callbacks.iteritems():
            if prop_name in all_properties:
                all_properties[prop_name]['callbacks'] = cb
            elif prop_name in self.event_dispatcher_event_callbacks:
                self.event_dispatcher_event_callbacks[prop_name] = callbacks[prop_name]

    @contextlib.contextmanager
    def temp_unbind_all(self, *prop_name):
        """
        Context manager to temporarily suspend dispatching of the listed properties or events. Assigning a different
        value to these properties or dispatching events inside the with statement will not dispatch the bindings.
        :param prop_name: property or event names to suspend
        """
        # Enter / With
        property_callbacks = {}
        event_callbacks = {}
        for name in prop_name:
            if name in self.event_dispatcher_properties:
                property_callbacks[name] = self.event_dispatcher_properties[name]['callbacks']
                self.event_dispatcher_properties[name]['callbacks'] = []
            if name in self.event_dispatcher_event_callbacks:
                event_callbacks[name] = self.event_dispatcher_event_callbacks[name]
                self.event_dispatcher_event_callbacks[name] = []
        # Inside of with statement
        yield None
        # Finally / Exit
        for name in prop_name:
            if name in property_callbacks:
                self.event_dispatcher_properties[name]['callbacks'] = property_callbacks[name]
            if name in event_callbacks:
                self.event_dispatcher_event_callbacks[name] = event_callbacks[name]


class PropertyEncoder(json.JSONEncoder):
    """
    Encoder that helps with the JSON serializing of properties. In particular, ObservableDict and ObservableList
    """

    def default(self, o):
        try:
            r = super(PropertyEncoder, self).default(o)
        except TypeError as e:
            if isinstance(o, ObservableList):
                return o.list
            elif isinstance(o, ObservableDict):
                return o.dictionary
            else:
                raise e
        return r





__all__ = ['EventDispatcherException', 'BindError', 'InvalidOptionError']


class EventDispatcherException(Exception):
    pass


class BindError(EventDispatcherException):
    pass


class InvalidOptionError(EventDispatcherException):
    def __init__(self, value, options):
        self.value = value
        self.options = options

    def __str__(self):
        return "'%s' is not one of the following allowed options: %s" % (self.value, [i for i in self.options])



__author__ = 'calvin'
from weakref import WeakSet

from . import Property


"""
Create a dictionary of conversion methods between US Standard and Metric distance unit systems.
"""
from_meter_conversions = {'km': 1 / 1000., 'm': 1, 'cm': 100., 'mm': 1000.,
                          'ft': 3.28084, 'yards': 1.09361, 'miles': 0.000621371, 'inches': 39.3701, 'in': 39.3701}
ConversionFactors = {}

for v, v_per_meter in from_meter_conversions.iteritems():
    # For each entry, add the conversion to and from meters
    fmt1 = "m_to_{}".format(v)
    inv1 = "{}_to_m".format(v)
    ConversionFactors[fmt1] = v_per_meter
    ConversionFactors[inv1] = 1. / v_per_meter

    for q, q_per_meter in from_meter_conversions.items():
        # for each entry, add the conversion to and from other entries
        v_per_q = v_per_meter / q_per_meter
        fmt2 = "{}_to_{}".format(q, v)
        inv2 = "{}_to_{}".format(v, q)
        ConversionFactors[fmt2] = v_per_q
        ConversionFactors[inv2] = 1. / v_per_q


class UnitProperty(Property):
    unit_properties = WeakSet()

    def __init__(self, default_value, units):
        super(UnitProperty, self).__init__(default_value, units=units)

    @staticmethod
    def get_units(dispatcher, property_name):
        return dispatcher.event_dispatcher_properties[property_name]['units']

    @staticmethod
    def convert_to(dispatcher, property_name, units):
        info = dispatcher.event_dispatcher_properties[property_name]
        if info['units'] == units:
            return
        c = ConversionFactors["{}_to_{}".format(info['units'], units)]
        setattr(dispatcher, property_name, c * info['value'])
        info['units'] = units

    def register(self, instance, property_name, default_value):
        super(UnitProperty, self).register(instance, property_name, default_value)
        # Keep track of all the UnitProperties so that we can change them all when the unit system changes
        self.unit_properties.add(self)



__author__ = 'calvin'

from . import Property


class LimitProperty(Property):

    def __init__(self, default_value, min, max):
        super(LimitProperty, self).__init__(default_value, min=min, max=max)

    def __get__(self, obj, objtype=None):
        return obj.event_dispatcher_properties[self.name]['value']

    def __set__(self, obj, value):
        info = obj.event_dispatcher_properties[self.name]
        if value != info['value']:
            # Clip the value to be within min/max
            if value < info['min']:
                # Only dispatch if the current value is not already clipped to the minimum
                if info['value'] != info['min']:
                    info['value'] = info['min']
                else:
                    return
            elif value > info['max']:
                # Only dispatch if the current value is not already clipped to the maximum
                if info['value'] != info['max']:
                    info['value'] = value = info['max']
                else:
                    return
            else:
                info['value'] = value
            # Dispatch callbacks
            for callback in info['callbacks']:
                if callback(obj, value):
                    break

    def __delete__(self, obj):
        raise AttributeError("Cannot delete properties")

    @staticmethod
    def get_min(inst, name):
        return inst.event_dispatcher_properties[name]['min']

    @staticmethod
    def set_min(inst, name, value):
        inst.event_dispatcher_properties[name]['min'] = value
        setattr(inst, name, value)

    @staticmethod
    def get_max(inst, name):
        return inst.event_dispatcher_properties[name]['max']

    @staticmethod
    def set_max(inst, name, value):
        inst.event_dispatcher_properties[name]['max'] = value
        setattr(inst, name, value)



__author__ = 'calvin'

from copy import deepcopy

from eventdispatcher import Property
from weakref import ref


class WeakRefProperty(Property):
    """
    Property that stores it's values as weak references in order to facilitate garbage collection.
    """

    def __init__(self, default_value, **additionals):
        self.instances = {}
        try:
            self.default_value = ref(default_value)
        except TypeError:
            self.default_value = None
        try:
            self.value = ref(deepcopy(default_value))
        except TypeError:
            self.value = None
        self._additionals = additionals

    def __get__(self, obj, objtype=None):
        value = obj.event_dispatcher_properties[self.name]['value']
        if value:
            return value()
        else:
            return value

    def __set__(self, obj, value):
        wr = ref(value) if value is not None else None
        if wr != obj.event_dispatcher_properties[self.name]['value']:
            prop = obj.event_dispatcher_properties[self.name]
            prop['value'] = wr
            for callback in prop['callbacks']:
                if callback(obj, value):
                    break

    def register(self, instance, property_name, default_value):
        wr = None if default_value is None else ref(default_value)
        super(WeakRefProperty, self).register(instance, property_name, wr)


__author__ = 'calvin'

from copy import deepcopy


class Property(object):

    def __init__(self, default_value, **additionals):
        self.instances = {}
        self.default_value = default_value
        self.value = deepcopy(default_value)
        self._additionals = additionals

    def __get__(self, obj, objtype=None):
        return obj.event_dispatcher_properties[self.name]['value']

    def __set__(self, obj, value):
        if value != obj.event_dispatcher_properties[self.name]['value']:
            prop = obj.event_dispatcher_properties[self.name]
            prop['value'] = value
            for callback in prop['callbacks']:
                if callback(obj, value):
                    break

    def __delete__(self, obj):
        raise AttributeError("Cannot delete properties")

    def register(self, instance, property_name, default_value):
        info = self._additionals.copy()
        info.update({'property': self, 'value': default_value, 'name': property_name, 'callbacks': []})
        # Create the instances dictionary at registration so that each class has it's own instance of it.
        self.instances[instance] = info
        instance.event_dispatcher_properties[property_name] = info

    def get_dispatcher_property(self, property_name):
        return self.instances[self][property_name]





__author__ = 'calvin'
import collections
from functools import partial
from . import Property


class ObservableSet(collections.MutableSet):

    def __init__(self, dictionary, dispatch_method):
        self.set = dictionary.copy()
        self.dispatch = dispatch_method

    def __repr__(self):
        return self.set.__repr__()

    def __get__(self, instance, owner):
        return self.set

    def __contains__(self, item):
        return item in self.set

    def __len__(self):
        return len(self.set)

    def __getitem__(self, item):
        return self.set[item]

    def __setitem__(self, key, value):
        try:
            prev = self.set[key]
            check = prev != value
        except KeyError:
            check = True
        self.set[key] = value
        if check:
            self.dispatch(self.set)

    def __eq__(self, other):
        # Must be this order and not self.set == other, otherwise unittest.assertEquals fails
        return other == self.set

    def __cmp__(self, other):
        return self.set == other

    def __ne__(self, other):
        return self.set != other

    def __delitem__(self, key):
        del self.set[key]
        self.dispatch(self.set)

    def __iter__(self):
        return iter(self.set)

    def __nonzero__(self):
        return bool(self.set)

    def __getstate__(self):
        return self.set

    def __reduce__(self):
        return (set, (tuple(self.set),), None, None, None)

    def add(self, value):
        self.set.add(value)

    def discard(self, value):
        self.set.discard(value)

    def copy(self):
        return self.__class__(self.set, self.dispatch)

    def get(self, key, default=None):
        self.set.get(key, default)

    def remove(self, item):
        self.set.remove(item)
        self.dispatch(self.set)

    def update(self, *items):
        if self.set != items:
            self.set.update(*items)
            self.dispatch(self.set)

    def pop(self):
        item = self.set.pop()
        self.dispatch(self.set)
        return item

    def difference(self, items):
        return self.set.difference(items)


class SetProperty(Property):

    def __init__(self, default_value, **kwargs):
        super(SetProperty, self).__init__(default_value, **kwargs)
        if not isinstance(default_value, set):
            raise ValueError('SetProperty takes sets only.')

    def register(self, instance, property_name, value):
        self.value = ObservableSet(value, dispatch_method=partial(instance.dispatch, property_name, instance))
        super(SetProperty, self).register(instance, property_name, self.value)

    def __set__(self, obj, value):
        p = self.instances[obj]
        do_dispatch = p['value'] != value
        p['value'].set.clear()
        p['value'].set.update(value)          # Assign to the ObservableDict's value
        if do_dispatch:
            for callback in p['callbacks']:
                if callback(obj, value):
                    break


author = 'calvin'

import gettext
from eventdispatcher import Property

# The translation (gettext) function to be used
translator = None


def fake_translation(s):
    """
    A fake translation function to 'french' that can help you verify that you have tagged all text in the program
    """
    return 'Le %s' % s if s != '\n' else '\n'


class StringProperty(Property):
    observers = set()

    def __init__(self, default_value):
        super(StringProperty, self).__init__(default_value)
        self.translatables = set()
        if not isinstance(default_value, basestring):
            raise ValueError('StringProperty can only accepts strings.')

    def register(self, instance, property_name, default_value):
        super(StringProperty, self).register(instance, property_name, default_value)
        if isinstance(default_value, _):
            prop = instance.event_dispatcher_properties[property_name]
            prop.update({'_': default_value, 'obj': instance})
            StringProperty.observers.add(self.translate)
            self.translatables.add(instance)

    def __get__(self, obj, owner):
        value = obj.event_dispatcher_properties[self.name]['value']
        if isinstance(value, _):
            return _.translate(value)
        else:
            return value

    def __set__(self, obj, value):
        prop = obj.event_dispatcher_properties[self.name]
        if isinstance(value, _):
            # If it is tagged as translatable, register this object as an observer
            prop.update({'_': value, 'obj': obj})
            StringProperty.observers.add(self.translate)
            self.translatables.add(obj)
        else:
            if obj in self.translatables:
                self.translatables.remove(obj)
        if value != prop['value']:
            prop['value'] = value
            for callback in prop['callbacks']:
                if callback(obj, value):
                    break

    def translate(self):
        for obj in self.translatables:
            prop = obj.event_dispatcher_properties[self.name]
            prop['value'] = _.translate(prop['_'])
            for callback in prop['callbacks']:
                callback(prop['obj'], prop['value'])

    @staticmethod
    def remove_translation():
        """
        Remove the currently set translation function and return the language back to english (or default)
        """
        StringProperty.set_translation_function(None)

    @staticmethod
    def set_translation_function(func):
        """
        Set the translation function and dispatch all changes
        """
        global translator
        translator = func
        # Dispatch the changes to all the observers
        for callback in StringProperty.observers:
            callback()

    @staticmethod
    def load_fake_translation():
        """
        Load a fake translation function to that can help you verify that you have tagged all text in the program.
        Adds 'Le' to the beginning of every string.
        """
        StringProperty.set_translation_function(fake_translation)

    @staticmethod
    def switch_lang(domain, localedir=None, languages=None, class_=None, fallback=False, codeset=None):
        global translator
        # Create the translation class from gettext
        translation = gettext.translation(domain, localedir, languages, class_, fallback, codeset)
        translator = translation.ugettext

        # Dispatch the changes to all the observers
        for callback in StringProperty.observers:
            callback()


class _(unicode):
    """
    This is a wrapper to the gettext translation function _(). This wrapper allows the eventdispatcher.StringProperty
    to be automatically updated when the language (translation function) changes. In this way, all labels will be
    re-translated automatically.

    """

    def __new__(cls, s, *args, **kwargs):
        if isinstance(s, _):
            s = unicode(s.untranslated)
        if translator:
            trans = translator(s, *args, **kwargs)
            obj = super(_, cls).__new__(cls, trans, *args, **kwargs)
        else:
            obj = super(_, cls).__new__(cls, s, *args, **kwargs)
        obj.untranslated = s
        return obj

    def __eq__(self, other):
        """
        Compare the fully translated string (including the _additionals) if comparing _ objects, otherwise compare
        the english strings
        """
        s = _.translate(self)
        if isinstance(other, _):
            return s == _.translate(other)
        else:
            return s == other

    def __ne__(self, other):
        """
        Compare the fully translated string (including the _additionals) if comparing _ objects, otherwise compare
        the english strings
        """
        s = _.translate(self)
        if isinstance(other, _):
            return s != _.translate(other)
        else:
            return s != other

    def __add__(self, other):
        """
        Rather than creating a new _ instance of the sum of the two strings, keep the added string as a reference so
        that we can translate each individually. In this way we can make sure the following translates correctly:

            eg.

                _('Show Lines') + '\n' + (_('On') if show_lines else _('Off'))

        """
        if not hasattr(self, '_additionals'):
            self._additionals = []

        if isinstance(other, _):
            self._additionals.append(other)
            if hasattr(other, '_additionals'):
                self._additionals.extend(other._additionals)
        else:
            self._additionals.append(other)

        return self

    def __mul__(self, other):
        if type(other) is bool:
            if other:
                return self
            else:
                return ''
        if type(other) is int:
            if not hasattr(self, '_additionals'):
                self._additionals = []
            self._additionals.extend([self] * other)
        else:
            raise TypeError("can't multiply sequence by non-int of type %s" % type(other))

    def __repr__(self):
        return u"{trans} ({orig})".format(trans=_.translate(self), orig=self.untranslated)

    def __str__(self):
        return _.translate(self)

    def __unicode__(self):
        return _.translate(self)

    @staticmethod
    def join_additionals(s):
        if translator is None:
            l = [s.untranslated]
            for a in s._additionals:
                l.append(a.untranslated if isinstance(a, _) else a)
        else:
            l = [translator(s.untranslated)]
            for a in s._additionals:
                l.append(translator(a.untranslated) if isinstance(a, _) else a)
        return ''.join(l)

    @staticmethod
    def translate(s):
        if isinstance(s, _):
            # If we were passed a translatable string object _
            if hasattr(s, '_additionals'):
                return _.join_additionals(s)
            else:
                if translator is None:
                    return s.untranslated
                else:
                    return translator(s.untranslated)
        else:
            if translator is None:
                return s
            else:
                return translator(s)

    @classmethod
    def join(cls, sep, iterable):
        for ii, s in enumerate(iterable):
            if ii == 0:
                t = cls(s)
            else:
                t += sep
                t += s
        try:
            return t
        except UnboundLocalError:
            return _('')



__author__ = 'calvin'
import collections
import json

from functools import partial
from . import Property


class ObservableDict(collections.MutableMapping):

    def __init__(self, dictionary, dispatch_method):
        self.dictionary = dictionary.copy()
        self.dispatch = dispatch_method

    def __repr__(self):
        return self.dictionary.__repr__()

    def __get__(self, instance, owner):
        return self.dictionary

    def __contains__(self, item):
        return item in self.dictionary

    def __getitem__(self, item):
        return self.dictionary[item]

    def __setitem__(self, key, value):
        try:
            prev = self.dictionary[key]
            check = prev != value
        except KeyError:
            check = True
        self.dictionary[key] = value
        if check:
            self.dispatch(self.dictionary)

    def __len__(self):
        return len(self.dictionary)

    def __eq__(self, other):
        return self.dictionary == other

    def __cmp__(self, other):
        return self.dictionary == other

    def __ne__(self, other):
        return self.dictionary != other

    def __delitem__(self, key):
        del self.dictionary[key]
        self.dispatch(self.dictionary)

    def __iter__(self):
        return iter(self.dictionary)

    def __nonzero__(self):
        return bool(self.dictionary)

    def __getstate__(self):
        return self.dictionary

    def __reduce__(self):
        return (dict, tuple(), None, None, self.dictionary.iteritems())

    def copy(self):
        return self.dictionary.copy()

    def get(self, key, default=None):
        return self.dictionary.get(key, default)

    def itervalues(self):
        return self.dictionary.itervalues()

    def iterkeys(self):
        return self.dictionary.iterkeys()

    def iteritems(self):
        return self.dictionary.iteritems()

    def update(self, E=None, **F):
        if E and self.dictionary != E:
            self.dictionary.update(E)
            self.dispatch(self.dictionary)
        elif F and self.dictionary != F:
            self.dictionary.update(F)
            self.dispatch(self.dictionary)

    def keys(self):
        return self.dictionary.keys()

    def values(self):
        return self.dictionary.values()

    def items(self):
        return self.dictionary.items()

    def pop(self, key):
        item = self.dictionary.pop(key)
        self.dispatch(self.dictionary)
        return item


class DictProperty(Property):

    def __init__(self, default_value, **kwargs):
        super(DictProperty, self).__init__(default_value, **kwargs)
        if not isinstance(default_value, dict):
            raise ValueError('DictProperty takes dict only.')

    def register(self, instance, property_name, value):
        self.value = ObservableDict(value, dispatch_method=partial(instance.dispatch, property_name, instance))
        super(DictProperty, self).register(instance, property_name, self.value)

    def __set__(self, obj, value):
        p = self.instances[obj]
        do_dispatch = p['value'] != value
        p['value'].dictionary.clear()
        p['value'].dictionary.update(value)          # Assign to the ObservableDict's value
        if do_dispatch:
            for callback in p['callbacks']:
                if callback(obj, value):
                    break


__author__ = 'calvin'

__author__ = 'calvin'

from time import time
import threading
import weakref
from clock import Clock


class ScheduledEvent(object):
    """ Creates a trigger to the scheduler generator that is thread-safe."""
    objs = weakref.WeakValueDictionary({})
    RUNNING = 1
    KILL = 0

    def __init__(self, func, timeout=0):
        self.func = func
        self.t0 = time()
        self.timeout = timeout
        self.lock = threading.Lock()
        self._active = 1
        ScheduledEvent.clock = Clock.get_running_clock()

    def __iter__(self):
        return self

    def stop(self):
        """
        Stops the scheduled event from being called.
        """
        self._active = 0

    def start(self):
        """
        Allows the scheduled event from being called again.
        """
        self._active = 1
        self.t0 = time()
        if self.clock.scheduled_funcs[self.next] == 0:
            self.clock.scheduled_funcs[self.next] += 1
            self.clock.queue.append(self.next)

    def kill(self):
        """
        Send a kill signal to the generator, kicking it out of it's while loop and closing the generator.
        We must catch the StopIteration exception so that the main loop does not fail.
        """

        def _kill():
            try:
                self.generator.send(ScheduledEvent.KILL)
            except StopIteration:
                pass
        # We need to schedule the kill, in case it is being called from within the function/generator
        self.clock.schedule_event(_kill)

    def reset_timer(self):
        """
        Reset the time reference, delaying any scheduled events (schedule_once, schedule_interval).
        """
        self.t0 = time()

        # schedule the call to the generator, ensuring only one function is added to the queue
        if not self.clock.scheduled_funcs[self.next]:
            self.clock.scheduled_funcs[self.next] += 1
            self.clock.queue.append(self.next)

    def reset_trigger(self, reschedule=False):
        if self.clock.scheduled_funcs[self.func]:
            self.clock.scheduled_funcs[self.func] -= 1
            self.clock.queue.remove(self.func)
        if reschedule:
            self.next()

    @property
    def is_scheduled(self):
        return bool(self.clock.scheduled_funcs[self.func])

    def __repr__(self):
        return "ScheduledEvent for {}{}".format(self.func, ' (scheduled)' if self.is_scheduled else '')

    def next(self, *args):
        try:
            with self.lock:
                self.generator.send(ScheduledEvent.RUNNING)
        except StopIteration:
            pass

    @staticmethod
    def unschedule_event(func):
        """
        Unschedule an event in the queue. Fails safely if the scheduled function is not in the queue.
        Be sure to use the same reference object if the scheduled function was a lambda or partial.
        :param func: scheduled function in the queue
        :return: True if the function was removed from the queue
        """
        clock = ScheduledEvent.clock
        if clock.scheduled_funcs[func]:
            clock.scheduled_funcs[func] -= 1
            clock.queue.remove(func)
            return True
        else:
            return False

    @staticmethod
    def schedule_once(func, timeout=0, start=True):
        """
        Schedule a function to be called `interval` seconds later.
        :param func: Scheduled Function
        :param interval: Time interval in seconds
        """
        if timeout:
            s = ScheduledEvent(func, timeout)
            s.generator = s._timeout_generator(func)
            s.generator.next()
            s.start() if start else s.stop()
            return s
        else:
            ScheduledEvent.clock.scheduled_funcs[func] += 1
            ScheduledEvent.clock.queue.append(func)

    @staticmethod
    def create_trigger(func):
        """
        Create a trigger that schedules a function to be called on the next cycle of the main loop.
        Calling the trigger more than once will not schedule the function multiple times.
        :param func: Scheduled Function
        """
        s = ScheduledEvent(func, timeout=0)
        s.generator = s._trigger_generator(func)
        s.generator.next()
        return s

    @staticmethod
    def schedule_interval(func, interval, start=False):
        """
        Schedule a function to be called every `interval` seconds. Must call start() to activate.
        :param func: Scheduled Function
        :param start: start right away.
        :param interval: Time interval in seconds
        """
        s = ScheduledEvent(func, timeout=interval)
        s.generator = s._interval_generator(func)
        s.generator.next()
        s.start() if start else s.stop()
        ScheduledEvent.objs[id(s)] = s
        return s

    """
    Generators for the different types of ScheduledEvents
    """

    def _interval_generator(self, f):
        """
        Generator. The function f is called every `timeout` number of seconds.
        """
        interval = self.timeout
        scheduled_funcs = self.clock.scheduled_funcs
        append = self.clock.queue.append
        _next = self.next
        running = yield
        while running:
            t = time()
            dt = t - self.t0
            if dt > interval and self._active:
                # If we have past the timeout time, call the function, reset the reference time (t0)
                f()
                self.t0 = t
            # Add another call to this generator to the scheduled functions
            if not scheduled_funcs[_next]:
                scheduled_funcs[_next] += 1
                append(_next)
            running = yield
        # When the loop breaks, we still have one scheduled call to the generator.
        yield

    def _timeout_generator(self, f):
        """
        Generator. The function f is called after `timeout` number of seconds
        """
        timeout = self.timeout
        scheduled_funcs = self.clock.scheduled_funcs
        append = self.clock.queue.append
        _next = self.next
        running = yield
        while running:
            dt = time() - self.t0
            if dt > timeout and self._active:
                # If we have pasted the timeout time, call the function
                f()
                running = yield
            else:
                # Add another call to this generator to the scheduled functions
                if not scheduled_funcs[_next]:
                    scheduled_funcs[_next] += 1
                    append(_next)
                running = yield
        # When the loop breaks, we still have one scheduled call to the generator.
        yield

    def _trigger_generator(self, f):
        """
        Generator. The function f is called on the next Clock cycle. A function can be scheduled at most once per
        clock cycle.
        """
        scheduled_funcs = self.clock.scheduled_funcs
        append = self.clock.queue.append
        running = yield
        while running:
            if not scheduled_funcs[f] and self._active:
                scheduled_funcs[f] += 1
                append(f)
                running = yield
            else:
                running = yield


__author__ = 'calvin'

import collections
import json
from functools import partial

from . import Property


class ObservableList(collections.MutableSequence):
    def __init__(self, l, dispatch_method):
        if not type(l) == list and not type(l) == tuple and not isinstance(l, ObservableList):
            raise ValueError('Observable list must only be initialized with lists as arguments')
        self.list = list(l)
        self.dispatch = dispatch_method

    def __repr__(self):
        return self.list.__repr__()

    def __get__(self, instance, owner):
        return self.list

    def __getitem__(self, item):
        return self.list[item]

    def __setitem__(self, key, value):
        if self.list[key] != value:
            self.list[key] = value
            self.dispatch(self.list)

    def __reversed__(self):
        return reversed(self.list)

    def __delitem__(self, key):
        del self.list[key]
        self.dispatch(self.list)

    def __len__(self):
        return len(self.list)

    def __iter__(self):
        return iter(self.list)

    def __nonzero__(self):
        return bool(self.list)

    def __getstate__(self):
        return self.list

    def __reduce__(self):
        return (list, tuple(), None, iter(self.list), None)

    def insert(self, index, value):
        self.list.insert(index, value)
        self.dispatch(self.list)

    def append(self, value):
        self.list.append(value)
        self.dispatch(self.list)

    def extend(self, values):
        self.list.extend(values)
        self.dispatch(self.list)

    def pop(self, index=-1):
        value = self.list.pop(index)
        self.dispatch(self.list)

        return value

    def __eq__(self, other):
        return self.list == other

    def __ne__(self, other):
        return self.list != other

    def __nonzero__(self):
        return bool(self.list)


class ListProperty(Property):

    def register(self, instance, property_name, value):
        self.value = ObservableList(value, dispatch_method=partial(instance.dispatch, property_name, instance))
        super(ListProperty, self).register(instance, property_name, self.value)

    def __set__(self, obj, value):
        p = self.instances[obj]
        # Check if we need to dispatch
        do_dispatch = len(p['value'].list) != len(value) or not ListProperty.compare_sequences(p['value'], value)
        self.instances[obj]['value'].list[:] = value        # Assign to ObservableList's value
        if do_dispatch:
            for callback in p['callbacks']:
                if callback(obj, value):
                    break

    @staticmethod
    def compare_sequences(iter1, iter2):
        """
        Compares two iterators to determine if they are equal. Used to compare lists and tuples
        """
        iter1, iter2 = iter(iter1), iter(iter2)
        for i1 in iter1:
            try:
                i2 = next(iter2)
            except StopIteration:
                return False

            if i1 != i2:
                return False

        try:
            i2 = next(iter2)
        except StopIteration:
            return True

        return False


__author__ = 'lobocv'

from distutils.core import setup
from eventdispatcher import __version__


setup(
    name='eventdispatcher',
    packages=['eventdispatcher'],
    version=__version__,
    description="An event dispatcher framework inspired by the Kivy project.",
    author='Calvin Lobo',
    author_email='calvinvlobo@gmail.com',
    url='https://github.com/lobocv/eventdispatcher',
    download_url='https://github.com/lobocv/eventdispatcher/tarball/%s' % __version__,
    keywords=['event', 'dispatcher', 'dispatching', 'kivy', 'observer', 'framework', 'property', 'properties'],
    classifiers=[],
)


__author__ = 'calvin'


from collections import deque, Counter


class Clock(object):

    clock = None

    def __init__(self, *args, **kwargs):
        self.scheduled_funcs = Counter()
        self.queue = deque([])
        self._running = 0
        Clock.clock = self
        super(Clock, self).__init__(*args, **kwargs)

    @staticmethod
    def get_running_clock():
        return Clock.clock

    def _run_scheduled_events(self):
        events = self.queue
        funcs = self.scheduled_funcs
        popleft = events.popleft
        for i in xrange(len(events)):
            f = popleft()
            funcs[f] -= 1
            f()

    def run(self):
        # Use all local variables to speed up the loop
        _run_scheduled_events = self._run_scheduled_events
        self._running = 1
        while self._running:
            _run_scheduled_events()



from . import Property
from .exceptions import InvalidOptionError


class OptionProperty(Property):

    def __init__(self, default_value, options, handler=None):
        super(OptionProperty, self).__init__(default_value, options=options, handler=None)
        self.handler = handler
        self.options = set(options) if options else set()
        if default_value not in self.options:
            raise ValueError('Default value must be one of the defined options.')

    def __set__(self, obj, value):
        if value in self.options:
            super(OptionProperty, self).__set__(obj, value)
        elif self.handler:
            self.handler(obj, value)
        else:
            raise InvalidOptionError(value, self.options)

    @staticmethod
    def set_options(inst, name, options):
        inst.event_dispatcher_properties[name]['options'].options = set(options)

    @staticmethod
    def set_handler(inst, name, handler):
        inst.event_dispatcher_properties[name]['handler'].handler = handler